// fine.C  Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 10:55:22 tac>
//
// 色々な形式のテキストファイルをdansデータファイル形式に変換する。
// 書式付き入力と、フリーフォーマット形式の入力を行うことができる。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W fine fine.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "fine  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:fine [<OPTION> ...] [<DATA FILE ...>]",
  "  Convert various text files to DANS data files.",
  "<OPTION>",
  "  -d[=<DELIMITER>]:Specify the data delimiter",
  "    -d:space(default)  -d=\\t:tab",
  "  -f=<FILE>:reads convert formulas or options from the specified file",
  "  -fint[={-|+}]:specify whether integerized data should be converted to",
  "    real numbers",
  "    -:not convert(default)  +:convert  -fint:-fint=+",
  "  -fmt=[<FORMAT>[,<FORMAT>,...]]:Specify the data conversion format",
  "    -fmt=:format is disable(-d option is enabled, default)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -item[={-|+}]:specify whether the first line of the file should be",
  "    converted to an item name",
  "    -:not convert  +:convert(default)  -item:-item=+"
};

using lformat = list<format>;   // 書式指定リストの別名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vdelm;                 // データの区切り文字列
  bool vfint;                   // データ整数化フラッグ
  bool vitem;                   // 項目名変換フラッグ
  lformat vfmtlst;              // 書式指定リスト
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const string& delm() const    // データの区切り文字列を返す
  { return vdelm; }
  bool fint() const             // データ整数化フラッグを返す
  { return vfint; }
  bool item() const             // 項目名変換フラッグを返す
  { return vitem; }
  const lformat& fmtlst() const // 書式指定リストを返す
  { return vfmtlst; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vdelm(" "), vfmtlst(), vfint(false), vitem(true)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "d")            // 区切り文字指定
  {
    vdelm = " ";  sstrtostr(opt[1]);
    if (!(opt[1].empty())) vdelm = opt[1];
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "fint")    // 整数化変換指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vfint = true;
    else if (opt[1] == "-") vfint = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "fmt")     // 変換書式指定
  {
    if (opt[1].empty()) vfmtlst.clear(); // 変換指定をクリアする
    else if (getfmtlst(vfmtlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "item")    // 項目名変改指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vitem = true;
    else if (opt[1] == "-") vitem = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

class finclass                  // finメインクラス
{
private:
  string fname;                 // ファイル名
  int get_data(const option& opt, datac& cdata, const string& str) const;
  // 文字列からデータを取り出し、取り出したデータ数を返す
  strsize_t get_subdata(const option& opt, string& substr, const string& sstr, const format& fmt) const;
  // 書式指定に従って文字列からデータを取り出す
public:
  finclass() : fname()          // コンストラクタ
  {}
  ~finclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込みメモリに保存する
  void put(idd& id, ldata& datalst) const; // データを出力する
};

int finclass::get_data(const option& opt, datac& cdata, const string& sstr) const
// 文字列からデータを取り出し、取り出したデータ数を返す
{
  if (sstr.empty()) return 0;   // 空文字列
  if ((opt.fmtlst()).empty()) return cdata.getdata(sstr, (opt.delm())[0]);
  string tstr(sstr), substr;
  lstr tstrlst;
  lformat::const_iterator pos = (opt.fmtlst()).begin();
  lformat::const_iterator pospre = pos;
  strsize_t idx = string::npos;
  while ((idx = get_subdata(opt, substr, tstr, *pos)) != string::npos)
  {
    if ((pos->str()).empty()) tstrlst.push_back(substr); // /文字列/形式は無視する
    if (idx < tstr.size()) tstr.erase(0, idx);
    else break;                 // 文字列終了
    pospre = pos;
    ++pos;
    if (pos == (opt.fmtlst()).end()) pos = pospre;
  }
  if (cdata.size() < 1) cdata.resize(tstrlst.size());
  size_t i= 0;
  lstr_citr lspos = tstrlst.begin();
  while(i < cdata.size() && lspos != tstrlst.end())
  {
    cdata.assign(i, *lspos);
    ++i;  ++lspos;
  }
  while (i < cdata.size())      // 不足分はブランクにする
  {
    cdata.assign(i, BLANKs);
    ++i;
  }
  return min(cdata.size(), tstrlst.size()); // 取り出したデータ数を返す
}

strsize_t finclass::get_subdata(const option& opt, string& substr, const string& sstr, const format& fmt) const
// 書式指定に従って文字列からデータを取り出す
{
  if (sstr.empty()) return string::npos;
  int datalen = abs(fmt.n()), n = 0, nc = 0;
  strsize_t idx = 0;
  substr.erase();
  string buf;
  while (n < datalen && idx < sstr.size())
  {
    uchar c = sstr[idx];
    if (nc > 1)                 // マルチバイトコードの途中
    { buf += c;  --nc; }
    else if (nc == 1)           // マルチバイトコードの最後
    {
      buf += c;
      substr += buf;
      if (ishkana(buf.c_str())) ++n; // 半角カタカナは1桁扱い
      else n += 2;                   // 全角文字は2桁扱い
      buf.erase();  nc = 0;
    }
    else if (ismcode(c))        // マルチバイトコード1バイト目
    { buf = c;  nc = ncode(c) - 1; }
    else                        // 1バイト文字
    { substr += c;  ++n; }
    ++idx;
  }
  strtrm(substr);               // 前後の空白を削除する
  if ((fmt.str()).empty())      // 通常の書式
  {
    double fdata = strtodf(substr);
    if (fmt.f() >= 0 && fdata != BLANKf) // 数字データ
    {
      if (opt.fint()) fdata /= pow(10.0, double(fmt.f()));
      ostringstream ostr;
      ostr << fdata;
      substr = ostr.str();
    }
  }
  return idx;                   // 次のデータの開始インデックスを返す
}

int finclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込みメモリに保存する
{
  istream* finp = nullptr;
  if (c_fname == nullptr)
  { finp = &cin;  c_fname = "stdin"; }
  else
  {
    finp = new ifstream(c_fname);
    if (!(*finp)) throw runtime_error(string("No such file->") + string(c_fname));
  }
  if (fname.empty()) fname = c_fname;
  string sline;                 // データ入力用
  datac cdata(id.ni());         // データオブジェクト
  while (getsline(sline, *finp) != EOF)
  {
    strchomp(sline);
    if (id.nc() < 1)            // 最初の読み込み
    {
      size_t ni = get_data(opt, cdata, sline);
      if (ni < 1) throw runtime_error(string("No data on the first line->") + sline);
      else id.set(fname, ni);
      if (opt.item())           // 1行目を項目として読み込む
      { for (size_t i = 0; i < ni; ++i) id.assign_item(i, cdata.sdata(i)); }
      else
      { datalst.push_back(cdata); }
    }
    else                        // 2行目以降の読み込み
    {
      get_data(opt, cdata, sline);
      datalst.push_back(cdata);
    }
    id.nc(id.nc() + 1);
  }
  if (finp != &cin) delete finp;
  return (int)(id.nc());
}

void finclass::put(idd& id, ldata& datalst) const // データを出力する
{
  id.put(cout, formtag);
  for (ldata_itr dpos = datalst.begin(); dpos != datalst.end(); ++dpos) dpos->put(cout);
}

int main(int argc, char* argv[]) // メインルーチン
{
  finclass fin;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; i++)
    {
      if (i >= argc)
      {
        if (fok) fin.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { fin.get(argv[i], opt, id, datalst);  fok = false; }
    }
    fin.put(id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
