// foute.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 10:57:57 tac>
//
// dansデータファイルをリスト形式で出力する。
// また、データを他のシステムのテキストファイル形式に変換することもできる。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W foute foute.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "foute  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:foute [<OPTION> ...] [<DATA FILE ...>]",
  "  Output a formatted list of DANS datas.",
  "<OPTION>",
  "  -b=<STRING>:converts a null string to the specified string and outputs it",
  "  -c=[<START NUMBER>][-[<END NUMBER>]]:specify case number to be output",
  "  -cd[={-|+}]:specify whether to output a case data",
  "    -:no output  +:output(default)  -cd:-cd=+",
  "  -d[=<DELIMITER>]:Specify the data delimiter",
  "    -d:space(default)  -d=\\t:tab",
  "  -f=<FILE>:reads options from the specified file",
  "  -fint[={-|+}]:specify whether integerized data should be converted to",
  "    real numbers",
  "    -:not convert(default)  +:convert  -fint:-fint=+",
  "  -fmt=[<FORMAT>[,<FORMAT>,...]]:Specify the data conversion format",
  "    -fmt=:format is disable(-d option is enabled, default)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -i=[<START NUMBER>][-[<END NUMBER>]][,...]:specify the output item",
  "    with number",
  "  -id[={-|+}]:specify whether to output a data ID information",
  "    -:no output  +:output(default)  -id:-id=+",
  "  -n[=<NUMBER OF DIGITS>]:output case number at the left end of case datas",
  "  -r<ITEM NUMBER>=[<LOWER LIMIT>][-[<UPPER LIMIT>]]:specify the lower and",
  "    upper limit for stratify items",
  "  -s<ITEM NUMBER>=<STRING>[,<STRING>,...]:specify the strings for stratify",
  "    items",
  "    output only cases that match the specified numeric range or strings",
  "    -r and -s can be mixed, and they are connected OR",
  "  -w={A4|B4|crt|<LINE LENGTH>}:specify the output width of line"
};

constexpr int Wmax = 256;       // 最大出力幅
constexpr int Wa4  = 80;        // A4用紙縦用出力幅
constexpr int Wb4  = 124;       // B4用紙横用出力幅
constexpr int Wcrt = 78;        // CRT用出力幅
constexpr int Ndef = 4;         // ケース番号の省略時桁数

using lformat = list<format>;   // 書式リストの別名

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vblank;                // ブランク文字列
  pairint vc;                   // 出力ケース番号
  bool vcd;                     // ケースデータ出力フラッグ
  string vdelm;                 // データの区切り文字列
  bool vfint;                   // データ整数化フラッグ
  lformat vfmtlst;              // 書式指定リスト
  list<pairint> vinolst;        // 出力項目番号リスト
  bool vid;                     // データ識別情報出力フラッグ
  int vn;                       // ケース番号の桁数
  int vw;                       // 1行の出力幅
  lstra vstralst;               // 層別情報リスト
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const string& blank() const   // ブランク文字列を返す
  { return vblank; }
  int cs() const                // 出力開始ケース番号を返す
  { return vc.first; }
  int ce() const                // 出力終了ケース番号を返す
  { return vc.second; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  const string& delm() const    // データの区切り文字列を返す
  { return vdelm; }
  bool fint() const             // データ整数化フラッグを返す
  { return vfint; }
  const lformat& fmtlst() const // 書式指定リストを返す
  { return vfmtlst; }
  const list<pairint>& inolst() const // 出力項目番号リストを返す
  { return vinolst; }
  bool id() const               // データ識別情報出力フラッグを返す
  { return vid; }
  int n() const                 // ケース番号の桁数を返す
  { return vn; }
  int w() const                 // 1行の出力幅を返す
  { return vw; }
  lstra& stralst()              // 層別情報リストを返す
  { return vstralst; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  bool ok(size_t ni) const     // オプション指定の整合性をチェックする
  { return straok(vstralst, ni); }
  bool smatch(datac& cdata) const // 層別のチェック
  { return stramatch(vstralst, cdata); }
};

option::option(const char* c_pname) : vpath(), vpname(), vblank(BLANKs), vc(1, INT_MAX), vcd(true), vdelm(), vfmtlst(), vfint(false), vinolst(), vid(true), vn(Ndef), vw(Wa4), vstralst()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "b")            // ブランク文字列指定
  {
    vblank = BLANKs;
    if (!(opt[1].empty())) vblank = strtodstr(opt[1]);
  }
  else if (opt[0] == "c")       // 出力ケース番号指定
  {
    if (getrange(vc.first, vc.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE, (double)1, (double)INT_MAX);
  }
  else if (opt[0] == "cd")      // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "d")       // 区切り文字指定
  {
    sstrtostr(opt[1]);
    if (!(opt[1].empty())) vdelm = opt[1];
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "fint")    // 整数化変換指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vfint = true;
    else if (opt[1] == "-") vfint = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "fmt")     // 変換書式指定
  {
    if (opt[1].empty()) vfmtlst.clear(); // 変換指定をクリアする
    else if (getfmtlst(vfmtlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "i")       // 出力項目の指定
  {
    if (getrangelst(vinolst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "n")       // ケース番号の出力桁指定
  {
    char* ptr = nullptr;
    if (opt[1].empty()) vn = Ndef;
    else vn = strtol((opt[1]).c_str(), &ptr, 0);
    if (vw - abs(vn) < 2 || (ptr != 0 && *ptr != '\0')) err = badopt(vpname, optstr, USAGE, 2 - vw, vw - 2);
  }
  else if (opt[0][0] == 'r' || opt[0][0] == 's') // 層別指定
  {
    stra tstra;
    char* prt = nullptr;
    int ino = strtol(opt[0].substr(1).c_str(), &prt, 0);
    if (tstra.get(opt[0][0], ino, opt[1]) == datanon) err = badopt(vpname, optstr, USAGE);
    if (!err) vstralst.push_back(tstra);
  }
  else if (opt[0] == "w")       // データの出力桁指定
  {
    char* ptr = nullptr;
    if (opt[1] == "A4") vw = Wa4;
    else if (opt[1] == "B4") vw = Wb4;
    else if (opt[1] == "crt") vw = Wcrt;
    else vw = strtol((opt[1]).c_str(), &ptr , 0);
    if (vw < 1 || Wmax < vw) err = badopt(vpname, optstr, USAGE, 1, Wmax);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

class foutclass                 // foutメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, const option& opt, idd& id) const;
  // データ識別情報を出力する
  void put_case(ostream& fout, const option& opt, const idd& id, int cno, datac& cdata, const vector<format>& fmtvec) const;
  // ケースデータを出力する
  void put_ffcase(ostream& fout, const option& opt, const idd& id, int nhed, datac& cdata, const vector<format>& fmtvec) const;
  // ケースデータをフリーフォーマットで出力する
  void put_fmtcase(ostream& fout, const option& opt, const idd& id, int nhed, datac& cdata) const;
  // ケースデータをフォーマット出力する
  int put_data(ostream& fout, const option& opt, const format* fmtp, int nhed, int widx, bool putdelm, const string& sdata) const;
  // 1つのデータを出力する
public:
  foutclass() : fname()         // コンストラクタ
  {}
  ~foutclass()                  // デストラクタ
  {}
  int get(const char* c_fname, const option& opt, idd& id, ldata& datalst);
  // データを読み込みメモリに保存する
  void put(option& opt, idd& id, ldata& datalst) const; // データを出力する
};

void foutclass::put_id(ostream& fout, const option& opt, idd& id) const
// データ識別情報を出力する
{
  formtype formt = (opt.n() == 0) ? formtag : formnon;
  id.put_title(fout, formt);
  id.put_stra(fout, formt);
  if ((opt.inolst()).empty())
  { id.put_item(fout, formt); }
  else
  {
    for (list<pairint>::const_iterator iitr = (opt.inolst()).begin(); iitr != (opt.inolst()).end(); ++iitr)
    {
      int istart = max(iitr->first, 1);
      int iend = (iitr->second <= (int)(id.ni())) ? iitr->second : id.ni();
      if (istart < iend)    // 正順出力
      {
        for (int ino = istart; ino <= iend; ++ino) id.put_item(fout, ino - 1, formt);
      }
      else                  // 逆順出力
      {
        for (int ino = istart; ino >= iend; --ino) id.put_item(fout, ino - 1, formt);
      }
    }
  }
  if (formt == formnon) fout << endl;
}

void foutclass::put_case(ostream& fout, const option& opt, const idd& id, int cno, datac& cdata, const vector<format>& fmtvec) const
// ケースデータを出力する
{
  int nhed = 0;                 // ケース番号の出力幅
  if (opt.n() != 0)             // ケース番号を出力する
  {
    nhed = abs(opt.n());
    if (opt.n() < 0) fout << left << setw(nhed) << cno << right;
    else fout << setw(nhed) << cno;
    fout << ":";  ++nhed;
  }
  if ((opt.fmtlst()).empty()) put_ffcase(fout, opt, id, nhed, cdata, fmtvec);
  else put_fmtcase(fout, opt, id, nhed, cdata);
}

void foutclass::put_ffcase(ostream& fout, const option& opt, const idd& id, int nhed, datac& cdata, const vector<format>& fmtvec) const
// ケースデータをフリーフォーマットで出力する
{
  int widx = nhed;
  bool putdelm = false;
  if ((opt.inolst()).empty())    // 項目指定がない場合
  {
    for (size_t i = 0; i < id.ni(); ++i)
    {
      const format* fmtp = (opt.delm().empty()) ? &(fmtvec[i]) : 0;
      widx = put_data(fout, opt, fmtp, nhed, widx, putdelm, (nhed > 0) ? cdata.sdata(i) : cdata.dstr(i));
      if (!putdelm) putdelm = true;
    }
  }
  else                          // 項目指定がある場合
  {
    for (list<pairint>::const_iterator iitr = (opt.inolst()).begin(); iitr != (opt.inolst()).end(); ++iitr)
    {
      int istart = max(0, iitr->first - 1);
      int iend = (iitr->second <= (int)(id.ni())) ? iitr->second - 1 : id.ni() - 1;
      if (istart < iend)
      {
        for (int i = istart; i <= iend; ++i)
        {
          const format* fmtp = (opt.delm().empty()) ? &(fmtvec[i]) : nullptr;
          widx = put_data(fout, opt, fmtp, nhed, widx, putdelm, (nhed > 0) ? cdata.sdata(i) : cdata.dstr(i));
          if (!putdelm) putdelm = true;
        }
      }
      else
      {
        for (int i = istart; i >= iend; --i)
        {
          const format* fmtp = (opt.delm().empty()) ? &(fmtvec[i]) : nullptr;
          widx = put_data(fout, opt, fmtp, nhed, widx, putdelm, (nhed > 0) ? cdata.sdata(i) : cdata.dstr(i));
          if (putdelm == false) putdelm = true;
        }
      }
    }
  }
  if (putdelm) fout << endl;
}

void foutclass::put_fmtcase(ostream& fout, const option& opt, const idd& id, int nhed, datac& cdata) const
// ケースデータをフォーマット出力する
{
  int widx = nhed;
  lformat::const_iterator itr = (opt.fmtlst()).begin(), itrpre = (opt.fmtlst()).begin();
  if ((opt.inolst()).empty())   // 項目指定がない場合
  {
    for (size_t i = 0; i < id.ni(); ++i)
    {
      while (itr != (opt.fmtlst()).end() && !((itr->str()).empty()))
      {                         // 書式指定中の文字列の出力
        widx = put_data(fout, opt, nullptr, nhed, widx, false, itr->str());
        itrpre = itr;
        ++itr;
      }
      if (itr == (opt.fmtlst()).end()) itr = itrpre; // 最後の書式を継続する
      widx = put_data(fout, opt, &(*itr), nhed, widx, false, cdata.sdata(i));
      itrpre = itr;
      ++itr;
    }
  }
  else                          // 項目指定がある場合
  {
    for (list<pairint >::const_iterator iitr = (opt.inolst()).begin(); iitr != (opt.inolst()).end(); ++iitr)
    {
      int istart = max(0, iitr->first - 1);
      int iend = (iitr->second <= (int)(id.ni())) ? iitr->second - 1 : id.ni() - 1;
      if (istart < iend)
      {
        for (int i = istart; i <= iend; ++i)
        {
          while (itr != (opt.fmtlst()).end() && !((itr->str()).empty()))
          {
            widx = put_data(fout, opt, nullptr, nhed, widx, false, itr->str());
            itrpre = itr;
            ++itr;
          }
          if (itr == (opt.fmtlst()).end()) itr = itrpre;
          widx = put_data(fout, opt, &(*itr), nhed, widx, false, cdata.sdata(i));
          itrpre = itr;
          ++itr;
        }
      }
      else
      {
        for (int i = istart; i >= iend; --i)
        {
          while (itr != (opt.fmtlst()).end() && !((itr->str()).empty()))
          {
            widx = put_data(fout, opt, nullptr, nhed, widx, false, itr->str());
            itrpre = itr;
            ++itr;
          }
          if (itr == (opt.fmtlst()).end()) itr = itrpre;
          widx = put_data(fout, opt, &(*itr), nhed, widx, false, cdata.sdata(i));
          itrpre = itr;
          ++itr;
        }
      }
    }
  }
  if (widx > 0) fout << endl;
}

int foutclass::put_data(ostream& fout, const option& opt, const format* fmtp, int nhed, int widx, bool putdelm, const string& sdata) const
// 1つのデータを出力する
{
  int dsize = sdata.size();
  int dwide = strclen(sdata, CHARwide);
  int npad = dsize - dwide;     // 文字列のバイト数と桁数の差
  if (putdelm && !((opt.delm()).empty()) && opt.delm() != "\0" && opt.delm() != "\n")
  { dwide += strclen(opt.delm(), CHARwide); }
  if (putdelm)
  {
    if (opt.delm() == "\n")
    { fout << "\n" << setw(nhed) << "";  widx = nhed; }
    else
    {
      if (nhed > 0 && dwide > opt.w() - widx)
      { fout << "\n" << setw(nhed) << "";  widx = nhed; }
      fout << opt.delm();
    }
  }
  double fdata = strtodf(sdata);
  if (fmtp == nullptr)          // 書式指定がない場合
  {
    if (sdata == "\n")          // 改行の場合(書式指定中の文字列)
    { fout << sdata << setw(nhed) << "";  widx = nhed; }
    else 
    {
      if (nhed > 0 && dwide > opt.w() - widx)
      { fout << "\n" << setw(nhed) << "";  widx = nhed; }
      fout << sdata;  widx += dwide;
    }
  }
  else if (fmtp->f() < 0 || fdata == BLANKf) // 文字データ
  {
    int w = max(abs(fmtp->n()), dwide);
    if (nhed > 0 && w > opt.w() - widx)
    { fout << "\n" << setw(nhed) << "";  widx = nhed; }
    if (fmtp->n() < 0) fout << left << setw(w + npad) << sdata << right;
    else fout << setw(w + npad) << sdata;
    widx += w;
  }
  else                          // 数字データ
  {
    int w = abs(fmtp->n());
    if (nhed > 0 && w > opt.w() - widx)
    { fout << "\n" << setw(nhed) << "";  widx = nhed; }
    int prec0 = fout.precision();
    fout.setf(ios::fixed);
    if (fmtp->n() < 0) fout << left;
    if (opt.fint() && fmtp->f() > 0) // 整数化数字データ
    {
      fdata *= pow(10.0, (double)(fmtp->f()));
      fout.precision(0);
    }
    else                        // 通常の数字データ
    { fout.precision(fmtp->f()); }
    fout << setw(w) << fdata;
    if (fmtp->n() < 0) fout << right;
    fout.precision(prec0);  fout.unsetf(ios::fixed);
    widx += w;
  }
  return widx;                  // 現在の桁数を返す
}

int foutclass::get(const char* c_fname, const option& opt, idd& id, ldata& datalst)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.ok(id.ni()) == false) throw runtime_error("Wrong specify of stratification !");
  return nc;
}

void foutclass::put(option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  vector<format> fmtvec(id.ni()); // 書式指定ベクトル
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    for (size_t i = 0; i < id.ni(); ++i)
    {
      if ((ditr->sdata(i) == BLANKs) && (opt.blank() != BLANKs)) ditr->assign(i, opt.blank());
      int w = strclen(ditr->sdata(i), CHARwide) + 1;
      if (w > fmtvec[i].n()) fmtvec[i].n(w);
    }
  }
  if (id.append_stra(opt.stralst()) < 0) throw runtime_error("Wrong specify of stratification !");
  if ((opt.inolst()).empty() == false) // 項目指定の正当性のチェック
  {
    for (list<pairint>::const_iterator itr = (opt.inolst()).begin(); itr != (opt.inolst()).end(); ++itr)
    {
      if (itr->first > (int)(id.ni()) || itr->second < 1)
      { throw runtime_error("Wrong specify of items !"); }
    }
  }
  if (opt.id()) put_id(cout, opt, id);
  if (opt.cd() == false) return;
  bool reverse = false;
  int cno = 1;
  if (opt.cs() > opt.ce())      // 逆順出力
  { reverse = true;  cno = datalst.size();  datalst.reverse(); }
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    if ((reverse == false && opt.cs() <= cno && cno <= opt.ce()) || (reverse && opt.cs() >= cno && cno >= opt.ce()))
    {
      if (opt.smatch(*ditr))
      {
        if (opt.cd()) put_case(cout, opt, id, cno, *ditr, fmtvec);
      }
    }
    if (reverse) --cno;
    else ++cno;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  foutclass fout;               // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) fout.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { fout.get(argv[i], opt, id, datalst);  fok = false; }
    }
    fout.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
