// merge.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 11:04:07 tac>
//
// 複数のdansのデータファイルをパラレルマージする。
// パラレルマージとは横方向のマージで、項目数が増え、
// ケース数は複数のファイルのうち最大のものになる。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -o merge merge.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "merge  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:merge [<OPTION> ...] <DATA FILE-1> [<DATA FILE-2> ...]",
  "  Parallel merge multiple data files.",
  "<OPTION>",
  "  -cd[={-|+}]:specify whether to output a case data",
  "    -:no output  +:output(default)  -cd:-cd=+",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -id[={-|+}]:specify whether to output a data ID information",
  "    -:no output  +:output(default)  -id:-id=+"
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vblank;                // ブランク文字列
  bool vcd;                     // ケースデータ出力フラッグ
  bool vid;                     // データ識別情報出力フラッグ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  bool id() const               // データ識別情報出力フラッグを返す
  { return vid; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vcd(true), vid(true)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cd")           // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

class filed                     // ファイルデータクラス
{
private:
  string fname;                 // ファイル名
  idd vid;                      // データ識別情報
  ldata vdatalst;               // データリスト
public:
  filed() : fname(), vid(), vdatalst() // デフォールトコンストラクタ
  {}
  filed(const filed& filed0) : fname(filed0.fname), vid(filed0.vid), vdatalst(filed0.vdatalst)
  // コピーコンストラクタ
  {}
  virtual ~filed()              // デストラクタ
  {}
  const filed& operator=(const filed& filed0); // コピー代入演算子
  int readf(const char* c_fname, const option& opt);
  // データを読み込みメモリに保存する
  idd& id()                     // データ識別情報を返す
  { return vid; }
  ldata& datalst()              // データリストを返す
  { return vdatalst; }
};

const filed& filed::operator=(const filed& filed0) // 代入演算子
{
  if (&filed0 == this) return *this; // 自分自身への代入を防ぐ
  fname = filed0.fname;  vid = filed0.vid;  vdatalst = filed0.vdatalst;
  return *this;
}

int filed::readf(const char* c_fname, const option& opt)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, vid, vdatalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (vid.ni() < 1) throw runtime_error(string("No such file !->") + fname);
  return nc;
}

using lfiled = list<filed>;     // ファイルデータリストの別名定義

class mergclass                 // mergメインクラス
{
public:
  mergclass()                   // コンストラクタ
  {}
  ~mergclass()                  // デストラクタ
  {}
  int get(const char* c_fname, const option& opt, lfiled& filedlst);
  // データを読み込みメモリに保存する
  void put(const option& opt, lfiled& filedlst) const;
  // データを出力する
};

int mergclass::get(const char* c_fname, const option& opt, lfiled& filedlst)
// データを読み込みメモリに保存する
{
  filed tfd;
  int nc = tfd.readf(c_fname, opt);
  filedlst.push_back(tfd);      // ファイルデータリストに追加する
  return nc;
}

void mergclass::put(const option& opt, lfiled& filedlst) const
// データを出力する
{
  lfiled::iterator fdpos = filedlst.begin();
  if (opt.id()) (fdpos->id()).put(cout, formtag); // データ識別情報出力
  size_t ncmax = (fdpos->id()).nc();
  ldata_itr* dpos = new ldata_itr[filedlst.size()];
  dpos[0] = (fdpos->datalst()).begin();
  ++fdpos;
  int k = 1;
  while (fdpos != filedlst.end())
  {
    if (opt.id()) (fdpos->id()).put_item(cout, formtag);
    if ((fdpos->id()).nc() > ncmax) ncmax = (fdpos->id()).nc();
    dpos[k] = (fdpos->datalst()).begin();
    ++fdpos;  ++k;
  }
  if (opt.cd() == false) return; // ケースデータ非出力
  for (size_t i = 0; i < ncmax; ++i)
  {
    bool putdelm = false;
    int k = 0;
    fdpos = filedlst.begin();
    while (fdpos != filedlst.end())
    {
      if (dpos[k] != (fdpos->datalst()).end())
      {
        for (size_t j = 0; j < (fdpos->id()).ni(); ++j)
        {
          if (putdelm) cout << CHRdelm;
          else putdelm = true;
          cout << (dpos[k])->dstr(j);
        }
        ++(dpos[k]);
      }
      else                      // ケースデータがない場合
      {
        for (size_t j = 0; j < (fdpos->id()).ni(); ++j)
        {
          if (putdelm) cout << CHRdelm;
          else putdelm = true;
          cout << BLANKs;
        }
      }
      ++fdpos;  ++k;
    }
    if (putdelm) cout << endl;
  }
}

int main(int argc, char *argv[]) // メインルーチン
{
  mergclass merg;               // メインクラス変数
  option opt(argv[0]);          // オプション
  lfiled filedlst;              // ファイルデータリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    for (int i = 1; i < argc; ++i) // 標準入力からは読み込まない
    {
      if (*(argv[i]) == CHRopt) opt.set(argv[i] + 1);
      else merg.get(argv[i], opt, filedlst);
    }
    merg.put(opt, filedlst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
