// sele.C  Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 09:51:34 tac>
//
// dansのデータファイルを入力し、項目・ケースなどを選択的に出力する。
// 新しいデータファイルを作成したり、
// 集計項目を選択して集計コマンドに送りこんだりするのに用いる。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W sele sele.C -I/usr/local/include
//              -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "sele  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:sele [<OPTION> ...] [<DATA FILE> ...]",
  "  Select and output data.",
  "<OPTION>",
  "  -b=<STRING>:converts a null string to the specified string and outputs it",
  "  -c=[<START NUMBER>][-[<END NUMBER>]]:specify case number to be output",
  "  -cd[={-|+}]:specify whether to output a case data",
  "    -:no output  +:output(default)  -cd:-cd=+",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -i=[<START NUMBER>][-[<END NUMBER>]][,...]:specify the output item",
  "    with number",
  "  -iname=<ITEM NAME>[,<ITEM NAME>,...]:specify the output item with name",
  "  -id[={-|+|i|s|t}{...}]:specify whether to output a data ID information",
  "    -:no output  +:output(default)  -id:-id=+",
  "    i:item name  s:information of stratify  t:title",
  "  -n[={-|+}]:specify whether to output a number of items and a number of cases",
  "    -:no output(default)  +:output  -n:-n=+",
  "  -r<ITEM NUMBER>=[<LOWER LIMIT>][-[<UPPER LIMIT>]]:specify the lower and",
  "    upper limit for stratify items",
  "  -s<ITEM NUMBER>=<STRING>[,<STRING>,...]:specify the strings for stratify",
  "    items",
  "    output only cases that match the specified numeric range or strings",
  "    -r and -s can be mixed, and they are connected OR"
};

using lino = list<pairint>;     // 項目番号リストの短名定義
using lino_itr = list<pairint>::iterator; // 反復子の短名定義
using lino_citr = list<pairint>::const_iterator; // const反復子の短名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vblank;                // ブランク文字列
  pairint vc;                   // 出力ケース番号
  bool vcd;                     // ケースデータ出力フラッグ
  lino vinolst;                 // 出力項目番号リスト
  lstr vinamelst;               // 出力項目名リスト
  bool vidt;                    // タイトル出力フラッグ
  bool vids;                    // 層別情報出力フラッグ
  bool vidi;                    // 項目名出力フラッグ
  bool vn;                      // 項目数・ケース数出力フラッグ
  lstra vstralst;               // 層別情報リスト
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const string& blank() const   // ブランク文字列を返す
  { return vblank; }
  int cs() const                // 出力開始ケース番号を返す
  { return vc.first; }
  int ce() const                // 出力終了ケース番号を返す
  { return vc.second; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  lino& inolst()                // 出力項目番号リストを返す
  { return vinolst; }
  lstr& inamelst()              // 出力項目名リストを返す
  { return vinamelst; }
  bool idt() const              // タイトル出力フラッグを返す
  { return vidt; }
  bool ids() const              // 層別情報出力フラッグを返す
  { return vids; }
  bool idi() const              // 項目名出力フラッグを返す
  { return vidi; }
  bool n() const                // 項目数・ケース数出力フラッグを返す
  { return vn; }
  lstra& stralst()              // 層別情報リストを返す
  { return vstralst; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  bool ok(size_t ni) const     // オプション指定の整合性をチェックする
  { return straok(vstralst, ni); }
  bool smatch(datac& cdata) const // 層別のチェック
  { return stramatch(vstralst, cdata); }
};

option::option(const char* c_pname) : vpath(), vpname(), vblank(BLANKs), vc(1, INT_MAX), vcd(true), vinolst(), vinamelst(), vidt(true), vids(true), vidi(true), vn(false), vstralst()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "b")            // ブランク文字列指定
  {
    vblank = BLANKs;
    if (!(opt[1].empty())) vblank = strtodstr(opt[1]);
  }
  else if (opt[0] == "c")       // 出力ケース番号指定
  {
    if (getrange(vc.first, vc.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE, 1, INT_MAX);
  }
  else if (opt[0] == "cd")      // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "i")       // 出力項目の指定
  {
    if (getrangelst(vinolst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty())
    { vidt = vids = vidi = true; }
    else
    {
      vidt = vids = vidi = false;
      for (size_t i = 0; i < opt[1].size(); ++i)
      {
        switch (opt[1][i])
        {
        case '-' : vidt = vids = vidi = false;
          break;
        case '+' : vidt = vids = vidi = true;
          break;
        case 't' : vidt = true;
          break;
        case 's' : vids = true;
          break;
        case 'i' : vidi = true;
          break;
        default : err = badopt(vpname, optstr, USAGE);
        }
        if (err) break;
      }
    }
  }
  else if (opt[0] == "iname")   // 出力項目名の指定
  {
    int n = strsplit(vinamelst, opt[1], CHRdelm);
    if (n < 1) err = badopt(vpname, optstr, USAGE);
    pairint inodm(INT_MAX, n); // ダミー項目番号
    vinolst.push_back(inodm);   // ここに項目番号を挿入する予定
  }
  else if (opt[0] == "n")       // 項目数・ケース数出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vn = true;
    else if (opt[1] == "-") vn = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0][0] == 'r' || opt[0][0] == 's') // 層別指定
  {
    stra tstra;
    char* prt = nullptr;
    int ino = strtol(opt[0].c_str() + 1, &prt, 0);
    if (tstra.get(opt[0][0], ino, opt[1]) == datanon) err = badopt(vpname, optstr, USAGE);
    vstralst.push_back(tstra);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

class selclass                  // selメインクラス
{
private:
  string fname;                 // ファイル名
  int set_ino(option& opt, idd& id) const; // 項目番号をセットして項目数を返す
  void put_id(ostream& fout, option& opt, idd& id, int nitem) const;
  // データ識別情報を出力する
  void put_case(ostream& fout, option& opt, const idd& id, datac& cdata) const;
  // ケースデータを出力する
public:
  selclass() : fname()          // コンストラクタ
  {}
  ~selclass()                   // デストラクタ
  {}
  int get(const char* c_fname, idd& id, ldata& datalst);
  // データを読み込みメモリに保存する
  void put(option& opt, idd& id, ldata& datalst) const; // データを出力する
};

int selclass::set_ino(option& opt, idd& id) const
// 項目番号をセットして項目数を返す
{
  if (opt.ok(id.ni()) == false)
  {
    disperr(opt.pname(), "Wrong -r or -s option !");
    return -1;
  }
  if (opt.cs() > 1 || (size_t)(opt.ce()) < id.nc()) // ケース番号指定あり
  {
    int ce = ((size_t)(opt.ce()) < id.nc()) ? opt.ce() : id.nc();
    ostringstream ostr;
    ostr << "Case Number=" << opt.cs() << "-" << ce;
    if (id.append_stra(ostr.str(), attnon, avalnon) < 0)
    {
      disperr(opt.pname(), "Wrong case number !");
      return -2;
    }
  }
  if (id.append_stra(opt.stralst()) < 0)
  {
    disperr(opt.pname(), "Wrong -r or -s option !");
    return -3;
  }
  if ((opt.inolst()).empty())   // 項目指定なし
  {
    pairint ino(1, id.ni());
    (opt.inolst()).push_back(ino); 
    return id.ni();             // 全項目出力
  }
  int nitem = 0;
  lstr_itr initr = (opt.inamelst()).begin();
  for (lino_itr iitr = (opt.inolst()).begin(); iitr != (opt.inolst()).end(); ++iitr)
  {
    if (iitr->first == INT_MAX) // 項目名による項目指定
    {
      int n = 0;
      pairint dino(0, 0);
      int niname = iitr->second;
      for (int k = 0; k < niname; ++k, ++initr)
      {
        string iname = *initr;
        strtrm(iname);          // 前後の空白を削除する
        sstrtostr(iname);       // 特殊表現文字を変換する(引用文字も削除)
        if (initr->size() < 1)
        {
          iname = CHRhead;
          iname += CHRtail;
        }
        else
        {
          if (iname[0] != CHRhead) iname = CHRhead + iname;
          if (iname[iname.size() - 1] != CHRtail) iname += CHRtail;
        }
        int ni = 0;
        for (size_t i = 0; i < id.ni(); ++i)
        {
          if (strpmatch(id.item(i), iname) > 0) // 項目名ヒット
          {                     // 2項目以降は要素を追加する
            if (n > 0) iitr = (opt.inolst()).insert(++iitr, dino);
            iitr->first = i + 1;  iitr->second = i + 1;
            ++n;  ++ni;
          }
        }
        if (ni < 1)
        {
          disperr(opt.pname(), "No such item name->", *initr);
          return -3;
        }
      }
      nitem += n;
    }
    else if ((size_t)(iitr->first) > id.ni() || iitr->second < 1)
    {
      disperr(opt.pname(), "Wrong specify for items !");
      return -4;
    }
    else
    {
      int istart = max(1, iitr->first);
      int iend = min(iitr->second, (int)(id.ni()));
      nitem += abs(iend - istart) + 1;
    }
  }
  return nitem;                 // 項目数を返す
}

void selclass::put_id(ostream& fout, option& opt, idd& id, int nitem) const
// データ識別情報を出力する
{
  if (opt.n())                  // データ識別情報の代わりに項目数出力
  { fout << nitem << "\n"; }
  else
  {
    if (opt.idt()) id.put_title(fout, formtag);
    if (opt.ids()) id.put_stra(fout, formtag);
    if (opt.idi())
    {
      for (lino_citr iitr = (opt.inolst()).begin(); iitr != (opt.inolst()).end(); ++iitr)
      {
        int istart = max(1, iitr->first);
        int iend = min(iitr->second, (int)(id.ni()));
        if (istart < iend)    // 正順出力
        {
          for (int ino = istart; ino <= iend; ++ino) id.put_item(fout, ino - 1, formtag);
        }
        else                  // 逆順出力
        {
          for (int ino = istart; ino >= iend; --ino) id.put_item(fout, ino - 1, formtag);
        }
      }
    }
  }
}

void selclass::put_case(ostream& fout, option& opt, const idd& id, datac& cdata) const
// ケースデータを出力する
{
  bool putdelm = false;
  for (lino_itr iitr = (opt.inolst()).begin(); iitr != (opt.inolst()).end(); ++iitr)
  {
    int istart = max(1, iitr->first) - 1;
    int iend = min(iitr->second, (int)(id.ni())) - 1;
    if (istart < iend)
    {
      for (int i = istart; i <= iend; ++i)
      {
        if (putdelm) fout << CHRdelm;
        else putdelm = true;
        if (cdata.sdata(i) == BLANKs) fout << opt.blank();
        else fout << cdata.dstr(i);
      }
    }
    else
    {
      for (int i = istart; i >= iend; --i)
      {
        if (putdelm) fout << CHRdelm;
        else putdelm = true;
        if (cdata.sdata(i) == BLANKs) fout << opt.blank();
        else fout << cdata.dstr(i);
      }
    }
  }
  if (putdelm) fout << endl;
}

int selclass::get(const char* c_fname, idd& id, ldata& datalst)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  return nc;                    // 読み込んだケース数を返す
}

void selclass::put(option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  int nitem = set_ino(opt, id);
  if (nitem < 0) return;
  if (opt.idt() || opt.ids() || opt.idi()) put_id(cout, opt, id, nitem);
  if (opt.cd() == false) return;
  bool reverse = false;
  int n = 0, cno = 1;
  if (opt.cs() > opt.ce())      // 逆順出力
  { reverse = true;  cno = datalst.size();  datalst.reverse(); }
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    if ((reverse == false && opt.cs() <= cno && cno <= opt.ce()) || (reverse && opt.cs() >= cno && cno >= opt.ce()))
    {
      if (opt.smatch(*ditr))
      {
        if (opt.cd() && !opt.n()) put_case(cout, opt, id, *ditr);
        ++n;
      }
    }
    if (reverse) --cno;
    else ++cno;
  }
  if (opt.cd() && opt.n()) cout << n << endl;
}

int main(int argc, char* argv[]) // メインルーチン
{
  selclass sel;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) sel.get(nullptr, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { sel.get(argv[i], id, datalst);  fok = false; }
    }
    sel.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
