// transe.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 11:07:04 tac>
//
// dansデータファイルの行(ケース)と列(項目)を入れかえる。
// この結果、元のファイルの項目数は出力ファイルのケース数になり、
// ケース数は項目数になる。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法:g++ -O -s -W transe transe.C
//             -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "transe  Version 7.41e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:sele [<OPTION> ...] [<DATA FILE> ...]",
  "  Replace rows(cases) and columns(items) in a data file.",
  "<OPTION>",
  "  -cd[={-|+}]:specify whether to output a case data",
  "    -:no output  +:output(default)  -cd:-cd=+",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -id[={-|+}]:specify whether to output a data ID information",
  "    -:no output  +:output(default)  -id:-id=+"
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vblank;                // ブランク文字列
  bool vcd;                     // ケースデータ出力フラッグ
  bool vid;                     // データ識別情報出力フラッグ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  bool id() const               // データ識別情報出力フラッグを返す
  { return vid; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vcd(true), vid(true)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cd")           // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

class transclass                // transメインクラス
{
private:
  string fname;                 // ファイル名
public:
  transclass() : fname()        // コンストラクタ
  {}
  ~transclass()                 // デストラクタ
  {}
  int get(const char* c_fname, idd& id, ldata& datalst);
  // データを読み込みメモリに保存する
  void put(const option& opt, idd& id, ldata& datalst) const;
  // データを出力する
};

int transclass::get(const char* c_fname, idd& id, ldata& datalst)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  return nc;
}

void transclass::put(const option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  if (opt.id())                 // データ識別情報出力
  {
    idd tid;
    tid.set(id.title(), id.nc());
    tid.put(cout, formtag);
  }
  if (opt.cd())                 // ケースデータ出力
  {
    for (size_t i = 0; i < id.ni(); ++i)
    {
      bool putdelm = false;
      for (ldata_itr dpos = datalst.begin(); dpos != datalst.end(); ++dpos)
      {
        if (putdelm) cout << CHRdelm;
        else putdelm = true;
        cout << dpos->dstr(i);
      }
      if (putdelm) cout << endl;
    }
  }
}

int main(int argc, char *argv[]) // メインルーチン
{
  transclass trans;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) trans.get(nullptr, id, datalst); // 標準入力から読み込む
      }
       else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { trans.get(argv[i], id, datalst);  fok = false; }
    }
    trans.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
